package bk;

import java.util.ArrayList;
import java.util.Stack;

import javax.swing.JTree;

import animation.AlgorithmStage;
import animation.AnimatedAlgorithm;
import animation.CodeLine;
import animation.ControlFlow;
import animation.Memory;
import animation.PseudoCodeNode;
import animation.PseudoCodeNode.CodeAction;
import animation.StackFrame;
import animation.StackFrame.FrameType;
import codelines.DeclareVariable;
import codelines.ForLoop;
import codelines.FunctionDefinition;
import graph.LayeredGraphEdge;
import graph.LayeredGraphNode;
import lib.TextLayoutHelper;

/**
 * the preprocessing stage of the bk node placement algorithm
 * 
 * @author kolja and eren
 *
 */
public class ConflictDetection implements AlgorithmStage {

    private LayeredGraphNode graph;
    private AnimatedAlgorithm alg;

    public ConflictDetection(LayeredGraphNode graph, AnimatedAlgorithm alg) {
        this.graph = graph;
        this.alg = alg;
    }

    @Override
    public PseudoCodeNode createPseudocodeTree(JTree tree) {
        String vars[] = { "i", "L", "k0", "l", "l1", "k1", "k", "v", "graph" };
        String params[] = { "graph" };
        PseudoCodeNode root = new PseudoCodeNode(TextLayoutHelper.setupPseudoCode("function mark_conflicts( graph )", vars), tree, new FunctionDefinition( params ), alg );
        PseudoCodeNode init = new PseudoCodeNode(TextLayoutHelper.setupPseudoCode( "L = graph.getContainedLayers();", vars), tree, new CodeLine() {
			@Override
			public ControlFlow runForward(Memory m) {
				m.declare( "L", m.<LayeredGraphNode>read( "graph", false ).getContainedLayers(), false );
				m.declare( "L.length-2", m.<LayeredGraphNode>read( "graph", false ).getContainedLayers().size() - 2, false );
				m.declare( "1", 1, false );
				actions.push( (Memory mem) -> {
					mem.undeclare( "L", false );
					mem.undeclare( "L.length-2", false );
					mem.undeclare( "1", false );
					return new ControlFlow( ControlFlow.STEP_OVER );
				} );
				return new ControlFlow( ControlFlow.STEP_OVER );
			}
        }, alg );
        root.add( init );
        PseudoCodeNode outerLoop = new PseudoCodeNode( TextLayoutHelper.setupPseudoCode( "for i=1 to |L|-2 do", vars ), tree, new ForLoop( "i", "1", "L.length-2" ), alg );
        root.add( outerLoop );
        PseudoCodeNode line = new PseudoCodeNode( TextLayoutHelper.setupPseudoCode( "k0 = 0; l = 0;", vars ), tree, new CodeLine() {
			@Override
			public ControlFlow runForward(Memory m) {
				m.declare( "k0", 0, false );
				m.declare( "l", 0, false );
				m.declare( "end_", m.<LayeredGraphNode>read( "graph", false ).getContainedLayers().get( m.<Integer>read( "i", false ) + 1 ).size() - 1, false );
				m.declare( "0", 0, false );
				actions.push( (Memory mem) -> {
					mem.undeclare( "k0", false );
					mem.undeclare( "l", false );
					mem.undeclare( "end_", false );
					mem.undeclare( "0", false );
					return new ControlFlow( ControlFlow.STEP_OVER );
				} );
				return new ControlFlow( ControlFlow.STEP_OVER );
			}
        }, alg );
        outerLoop.add( line );
        PseudoCodeNode innerLoop = new PseudoCodeNode( TextLayoutHelper.setupPseudoCode( "for l1=0 to |L[i+1]|-1 do", vars ), tree, new ForLoop( "l1", "0", "end_" ), alg );
        outerLoop.add( innerLoop );
        PseudoCodeNode ifNode = new PseudoCodeNode( TextLayoutHelper.setupPseudoCode( "if l1==|L[i+1]|-1 or L[i+1][l1] incident to inner segment between L[i+1] and L[i] then", vars ), tree, new CodeLine() {
			@Override
			public ControlFlow runForward(Memory m) {
				if( m.isDefined( "line_" + lineId + "_inside", false ) )
				{
					StackFrame old = m.removeFrame();
					actions.push( (Memory mem) -> {
						mem.addFrame( old );
		            	return new ControlFlow( ControlFlow.STEP_INTO );
					} );
	                return new ControlFlow( ControlFlow.STEP_OVER );
				}
				else
				{
		            if( m.<LayeredGraphNode>read( "graph", false ).getContainedLayers().get( m.<Integer>read( "i", false ) + 1).size() == m.<Integer>read( "l1", false ) || 
		            		incidentToInnerSegmentBetweenLiPlusOneAndLi( m ) ) {
		            	m.addFrame( new StackFrame( FrameType.LOOP ) );
		            	m.declare( "line_" + lineId + "_inside", true, false );
		            	actions.push( (Memory mem) -> {
		            		mem.removeFrame();
		            		return new ControlFlow( ControlFlow.STEP_OVER );
		            	} );
		            	return new ControlFlow( ControlFlow.STEP_INTO );
		            } else {
		            	m.declare( "line_" + lineId + "_inside", false, false );
		            	actions.push( (Memory mem) -> {
		            		mem.undeclare( "line_" + lineId + "_inside", false );
			                return new ControlFlow( ControlFlow.STEP_OVER );
		            	} );
		                return new ControlFlow( ControlFlow.STEP_OVER );
		            }
				}
			}
        }, alg );
        innerLoop.add( ifNode );
        /*
        lines[ 4 ] = new PseudoCodeNode( TextLayoutHelper.setupPseudoCode( "k1 = |L[i]|-1;", vars ), tree );
        lines[ 3 ].add( lines[ 4 ] );
        lines[ 5 ] = new PseudoCodeNode( TextLayoutHelper.setupPseudoCode( "if L[i+1][l1] incident to inner segment between L[i+1] and L[i] then", vars ), tree );
        lines[ 3 ].add( lines[ 5 ] );
        lines[ 6 ] = new PseudoCodeNode( TextLayoutHelper.setupPseudoCode( "k1 = pos(pred(L[i+1][l1])[0]);", vars ), tree );
        lines[ 5 ].add( lines[ 6 ] );
        lines[ 8 ] = new PseudoCodeNode( TextLayoutHelper.setupPseudoCode( "while l <= l1 do", vars ), tree );
        lines[ 3 ].add( lines[ 8 ] );
        lines[ 9 ] = new PseudoCodeNode( TextLayoutHelper.setupPseudoCode( "foreach v in pred(L[i+1][l]) do", vars ), tree );
        lines[ 8 ].add( lines[ 9 ] );
        lines[ 10 ] = new PseudoCodeNode( TextLayoutHelper.setupPseudoCode( "if k < k0 or k > k1 then mark segment (v,L[i+1][l]);", vars ), tree );
        lines[ 9 ].add( lines[ 10 ] );
        lines[ 12 ] = new PseudoCodeNode( TextLayoutHelper.setupPseudoCode( "l = l+1;", vars ), tree );
        lines[ 8 ].add( lines[ 12 ] );
        lines[ 14 ] = new PseudoCodeNode( TextLayoutHelper.setupPseudoCode( "k0 = k1;", vars ), tree );
        lines[ 3 ].add( lines[ 14 ] );
        lines[ 15 ] = new PseudoCodeNode( TextLayoutHelper.setupPseudoCodeStage( "mark conflicts in subgraphs" ), tree );
        root.add( lines[ 15 ] );*/
        return root;
    }
    
    private boolean incidentToInnerSegmentBetweenLiPlusOneAndLi( Memory m ) {
        LayeredGraphNode curr = m.<LayeredGraphNode>read( "graph", false ).getContainedLayers().get( m.<Integer>read( "i", false ) + 1 ).get( m.read( "l1", false ) );
        for (LayeredGraphEdge e : curr.getIncomingEdges()) {
            if (e.isDummyEdge()) {
                return true;
            }
        }
        return false;
    }

    @Override
    public String getDebugString() {/*
        String info = "| i  | l  | l1 | k0 | k1 | k  |\n";
        info +=       "|----|----|----|----|----|----|\n";
        info += "|" + TextLayoutHelper.strToLen( "" + i, 4 ) + 
                "|" + TextLayoutHelper.strToLen( "" + l, 4 ) + 
                "|" + TextLayoutHelper.strToLen( "" + l1, 4 ) + 
                "|" + TextLayoutHelper.strToLen( "" + k0, 4 ) + 
                "|" + TextLayoutHelper.strToLen( "" + k1, 4 ) + 
                "|" + TextLayoutHelper.strToLen( "" + k, 4 ) + "|\n";
        if( insideSubgraph && vIndex < graph.getContainedNodes().size() )
        {
            info += "Subgraph of " + graph.getContainedNodes().get( vIndex ).getName() + ":\n";
            String tmp = subgraphAlgs.get( vIndex ).getDebugString();
            info += tmp;
            return info;
        }
        return info;*/
    	return "";
    }

}
