package codelines;

import animation.CodeLine;
import animation.ControlFlow;
import animation.Memory;
import animation.StackFrame;
import animation.StackFrame.FrameType;

public class ForLoop extends CodeLine {

	String loopVar;
	private String start;
	private String end;
	
	public ForLoop( String varName, String start, String end )
	{
		this.loopVar = varName;
		this.start = start;
		this.end = end;
	}
	
	@Override
	public ControlFlow runForward(Memory m) {
		boolean declared = false; // prove if it is the first step in the loop
		if( !m.isSomewhereDefined( "line_" + lineId + "_index", false ) )
		{ // first loop step
			m.declare( "line_" + lineId + "_index", (int)m.<Integer>read( start, false ), false );
			declared = true;
		}
		if( m.<Integer>read( "line_" + lineId + "_index", false ) > m.<Integer>read( end, false ) ) // prove if the loop has finished
		{
			m.undeclare( "line_" + lineId + "_index", false );
			actions.push( (Memory mem) -> {
				return new ControlFlow( ControlFlow.STEP_OVER ); // loop was not called so nothing to reverse
			});
			return new ControlFlow( ControlFlow.STEP_OVER ); // don't execute the loop body
		}
		if( declared )
		{
			m.addFrame( new StackFrame( FrameType.LOOP ) );
			m.declare( loopVar, (int)m.<Integer>read( start, false ), false ); // set loop variable
			actions.push( (Memory mem) -> {
				mem.removeFrame();
				mem.undeclare( "line_" + lineId + "_index", false );
				return new ControlFlow( ControlFlow.STEP_OVER ); // step out of the loop
			} );
		}
		else
		{
			int oldIndex = m.<Integer>read( "line_" + lineId + "_index", false );
			m.write( "line_" + lineId + "_index", oldIndex + 1, false ); // count index down
			if( m.<Integer>read( "line_" + lineId + "_index", false ) > m.<Integer>read( end, false ) ) // prove if loop was finished
			{
				StackFrame sf = m.removeFrame(); // remove loop stack
				m.undeclare( "line_" + lineId + "_index", false );
				actions.add( (Memory mem) -> {
					mem.write( "line_" + lineId + "_index", oldIndex, false );
					mem.addFrame( sf ); // restore last loop stack
					return new ControlFlow( ControlFlow.STEP_INTO ); // step into the loop body
				});
				return new ControlFlow( ControlFlow.STEP_OVER ); // step out of the loop
			}
			StackFrame old = m.removeFrame(); // fresh stack frame for loop body
			m.addFrame( new StackFrame( FrameType.LOOP ) );
			m.write( loopVar, (int)m.read( "line_" + lineId + "_index", false ), false ); // update loop variable
			actions.push( (Memory mem) -> {
				mem.removeFrame();
				mem.addFrame( old );
				mem.write( "line_" + lineId + "_index", oldIndex, false );
				return new ControlFlow( ControlFlow.STEP_INTO );
			});
		}
		return new ControlFlow( ControlFlow.STEP_INTO );
	}
}
